/*
 * x86work library
 * 05.07.24 : Makoto NARA (Mc.N)
 */

/*
 * Copyright (c) 2005 Makoto NARA (Mc.N) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY Makoto NARA (Mc.N) ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "x86work.h"


BOOL
initBrandStrings(
	PCPUID_INFOR pci
	);

BOOL
initVendorID(
	PCPUID_INFOR pci
	);

PCPUIDS_DATA
searchCPUIDdata(
	PCPUIDS_DATA pcpus,
	DWORD cntcpu,
	DWORD idx
	);




BOOL
_cpuid(
	DWORD idx,
	PCPUID_DATA pcpu
	)
{
	return _cpuid_os(idx, pcpu);
}


BOOL
_rdmsr(
	DWORD idx,
	PMSR_DATA pmsr
	)
{
	return _rdmsr_os(idx, pmsr);
}




BOOL
CloseCPUIDinfo(
	PCPUID_INFOR pci
	)
{
	if( NULL == pci ){ return TRUE; }
	if( NULL != pci->pcpus ){
		CloseCPUIDdata(pci->pcpus);
		pci->pcpus = NULL;
	}
	free(pci);
	
	return TRUE;
}


BOOL
CloseCPUIDdata(
	PCPUIDS_DATA pcpus
	)
{
	PCPUIDS_DATA pnew, p = NULL;

	if( NULL == pcpus ){ return TRUE; }
	p = pcpus;
	for(;;){
		if( NULL == p->pnext ){ free(p); break; }
		pnew = p->pnext;
		free(p);
		p = pnew;
	}

	return TRUE;
}


PCPUID_INFOR
OpenCPUIDinfo(
	void
	)
{
	PCPUID_INFOR pci = NULL;
	BOOL bRet = FALSE;

	// get cpuid information structure
	pci = GetCPUIDdata_os();
	if( NULL == pci ){ return NULL; }

	// clear find first CPUID struct
	pci->pnext = NULL;
	// get vendor id
	initVendorID(pci);
	// get Brand Strings
	initBrandStrings(pci);

	return pci;
}

PCPUID_INFOR
PreLoadCPUIDinfo(
	void
	)
{
	PCPUID_INFOR pci = NULL;

	pci = (PCPUID_INFOR)malloc(sizeof(CPUID_INFOR));
	if( NULL != pci ){
		memset(pci, '\0', sizeof(CPUID_INFOR) );
	}
	return pci;
}

BOOL
PreAddCPUIDinfo(
	DWORD numcpu,
	DWORD cntcpu,
	DWORD ext,
	DWORD idx,
	DWORD eax,
	DWORD ebx,
	DWORD ecx,
	DWORD edx,
	PCPUID_INFOR pci
	)
{
	PCPUIDS_DATA p  =  NULL;
	PCPUIDS_DATA pt =  NULL;

	if( NULL == pci ){ return FALSE; }

	p = (PCPUIDS_DATA)malloc(sizeof(CPUIDS_DATA));
	if( NULL == p ){ return FALSE; }
	p->cpu.numcpu = numcpu;
	p->cpu.cntcpu = cntcpu;
	p->cpu.ext    = ext;
	p->cpu.idx    = idx;
	p->cpu.eax    = eax;
	p->cpu.ebx    = ebx;
	p->cpu.ecx    = ecx;
	p->cpu.edx    = edx;
	p->pnext      = NULL;

	if( NULL == pci->pcpus ){
		pci->pcpus = p;
	}else{
		for(pt = pci->pcpus; NULL != pt->pnext; pt = pt->pnext );
		pt->pnext = p;
	}

	if( pci->dwProcessor < cntcpu +1 ){
		pci->dwProcessor = cntcpu +1;
	}

	return TRUE;
}

BOOL
TreatCPUIDinfo(
	PCPUID_INFOR pci
	)
{
	if( NULL == pci ){ return FALSE; }
	// clear find first CPUID struct
	pci->pnext = NULL;
	// get vendor id
	initVendorID(pci);
	// get Brand Strings
	initBrandStrings(pci);

	return TRUE;
}

BOOL
initBrandStrings(
	PCPUID_INFOR pci
	)
{
	PCPUIDS_DATA pcpus = NULL;
	PBYTE pb = NULL;

	if( NULL==pci || NULL==pci->pcpus ){ return FALSE; }

	// init
	pci->szBrand[0] = '\0';
	pci->szBrand[MAX_BRANDSTRING-1] = '\0';
	pb = pci->szBrand;

	// search CPUID(EXTCPUID+4)
	if( NULL == (pcpus = searchCPUIDdata(pci->pcpus, 0, EXTCPUID+4)) ){ return FALSE; }
	*((DWORD*)(pb+32)) = pcpus->cpu.eax;
	*((DWORD*)(pb+36)) = pcpus->cpu.ebx;
	*((DWORD*)(pb+40)) = pcpus->cpu.ecx;
	*((DWORD*)(pb+44)) = pcpus->cpu.edx;
	// search CPUID(EXTCPUID+3)
	if( NULL == (pcpus = searchCPUIDdata(pci->pcpus, 0, EXTCPUID+3)) ){ return FALSE; }
	*((DWORD*)(pb+16)) = pcpus->cpu.eax;
	*((DWORD*)(pb+20)) = pcpus->cpu.ebx;
	*((DWORD*)(pb+24)) = pcpus->cpu.ecx;
	*((DWORD*)(pb+28)) = pcpus->cpu.edx;
	// search CPUID(EXTCPUID+2)
	if( NULL == (pcpus = searchCPUIDdata(pci->pcpus, 0, EXTCPUID+2)) ){ return FALSE; }
	*((DWORD*)(pb+0))  = pcpus->cpu.eax;
	*((DWORD*)(pb+4))  = pcpus->cpu.ebx;
	*((DWORD*)(pb+8))  = pcpus->cpu.ecx;
	*((DWORD*)(pb+12)) = pcpus->cpu.edx;

	return TRUE;
}




BOOL
initVendorID(
	PCPUID_INFOR pci
	)
{
	PCPUIDS_DATA pcpus = NULL;
	PBYTE pb = NULL;

	if( NULL==pci || NULL==pci->pcpus ){ return FALSE; }

	// init
	pci->dwVendor      = CPUVENDOR_UNKNOWN;
	pci->szVendorID[0] = '\0';
	pb = pci->szVendorID;

	// search CPUID(0)
	if( NULL == (pcpus = searchCPUIDdata(pci->pcpus, 0, 0)) ){ return FALSE; }

	*((DWORD*)(pb+0)) = pcpus->cpu.ebx;
	*((DWORD*)(pb+4)) = pcpus->cpu.edx;
	*((DWORD*)(pb+8)) = pcpus->cpu.ecx;
	*(pb+12) = '\0';

	if( 0 == strcmp(pb, "GenuineIntel") ){
		// INTEL
		pci->dwVendor = CPUVENDOR_INTEL;
	}else if( 0 == strcmp(pb, "AuthenticAMD") ){
		// AMD
		pci->dwVendor = CPUVENDOR_AMD;
	}else{
		// unknown...
		pci->dwVendor = CPUVENDOR_UNKNOWN;
	}

	return TRUE;
}




PCPUIDS_DATA
searchCPUIDdata(
	PCPUIDS_DATA pcpus,
	DWORD cntcpu,			// processor #
	DWORD idx				// CPUID(?)
	)
{
	PCPUIDS_DATA p = NULL;

	p = pcpus;

	for(;;){
		if( NULL==p ){ return NULL; }
		if( cntcpu == p->cpu.cntcpu && idx == p->cpu.idx ){
			return p;
		}
		p = p->pnext;
	}
	return NULL;
}



BOOL
IsFeature(
	PCPUID_INFOR pci,
	DWORD cntcpu,
	DWORD feature
	)
{
	PCPUIDS_DATA pcpus = NULL;
	BOOL bRet = FALSE;

	if( NULL==pci ){return FALSE;}

	switch(feature){
	case FLAG_CPUID_FPU:	// I`bv̕_Zjbg
	case FLAG_CPUID_VME:	// z[hg
	case FLAG_CPUID_DE:		// fobO@\g
	case FLAG_CPUID_PSE:	// y[WETCYg
	case FLAG_CPUID_TSC:	// ^CEX^vEJE^ (RDTSC)
	case FLAG_CPUID_MSR:	// fŗLWX^ (RDMSR/WRMSR)
	case FLAG_CPUID_PAE:	// AhXg (Ɍ PAE g)
	case FLAG_CPUID_MCE:	// }V`FbNO
	case FLAG_CPUID_CX8:	// CMPXCHG8
	case FLAG_CPUID_APIC:	// on-chip APIC
	case FLAG_CPUID_RSV1:	// reserved(intel)
	case FLAG_CPUID_SEP:	// VXeER[ (SYSENTER/SYSEXIT)
	case FLAG_CPUID_MTRR:	// E^CvEWEWX^
	case FLAG_CPUID_PGE:	// y[WEO[oECl[u
	case FLAG_CPUID_MCA:	// }VE`FbNEA[LeN`
	case FLAG_CPUID_CMOV:	// tړ߂̃T|[g (CMOV)
	case FLAG_CPUID_PAT:	// y[We[u
	case FLAG_CPUID_PSE36:	// 36-bit page size g
	case FLAG_CPUID_PSN:	// Processor Serial Number
	case FLAG_CPUID_CLFSH:	// CLFLUSH
	case FLAG_CPUID_RSV2:	// reserved(intel)
	case FLAG_CPUID_DS:		// Debug Store
	case FLAG_CPUID_ACPI:	// xj^у\tgEFA䂳NbN@\̃T|[g
	case FLAG_CPUID_MMX:	// MMX
	case FLAG_CPUID_FXSR:	// _ReLXg̍Z[u^XgA
	case FLAG_CPUID_SSE:	// SSE
	case FLAG_CPUID_SSE2:	// SSE2
	case FLAG_CPUID_SS:		// ZtEXk[v
	case FLAG_CPUID_HTT:	// Hyper-Threading
	case FLAG_CPUID_TM:		// xj^̃T|[g
	case FLAG_CPUID_RSV3:	// reserved(intel)
	case FLAG_CPUID_PBE:	// Pending Break
		// CPUID(1) and EDX (for Intel Standard)
		if( NULL == (pcpus = searchCPUIDdata(pci->pcpus, cntcpu, 1)) ){ return FALSE; }
		bRet = pcpus->cpu.edx&(1<<(feature&FLAG_CPUID_BITSHIFT)) ? TRUE : FALSE;
		break;
	case FLAG_CPUID_SSE3:		// SSE3
	case FLAG_CPUID_RSV4:		// reserved(intel)
	case FLAG_CPUID_RSV5:		// reserved(intel)
	case FLAG_CPUID_MONITOR:	// MONITOR/MWAIT
	case FLAG_CPUID_DS_CPL:		// CPL Qualified Debug Store
	case FLAG_CPUID_RSV6:	
	case FLAG_CPUID_RSV7:	
	case FLAG_CPUID_EST:		// Intel SpeedStep Technology
	case FLAG_CPUID_TM2:		// xj^Q
	case FLAG_CPUID_RSV8:	
	case FLAG_CPUID_CID:		// Context ID
	case FLAG_CPUID_RSV9:	
	case FLAG_CPUID_RSV10:	
	case FLAG_CPUID_CX16:		// CMPXCHG16B
	case FLAG_CPUID_xTPR:		// Send Task Priority Messages
		// CPUID(1) ECX (for Intel Standard)
		if( NULL == (pcpus = searchCPUIDdata(pci->pcpus, cntcpu, 1)) ){ return FALSE; }
		bRet = pcpus->cpu.ecx&(1<<(feature&FLAG_CPUID_BITSHIFT)) ? TRUE : FALSE;
		break;
	case FLAG_CPUID_SYSCALL:	// SYSCALL/SYSRET (Intel)
	case FLAG_CPUID_XDbit:		// NX-bit/XD-bit (AMD/Intel)
	case FLAG_CPUID_AMDMMX:		// AMD extensions MMX (AMD)
	case FLAG_CPUID_FFXSR:		// Fast FXSAVE and FXRSTOR (AMD)
	case FLAG_CPUID_EM64T:		// AMD64/EM64T (AMD/Intel)
	case FLAG_CPUID_3DNOWEXT:	// Extensions to 3DNow! (AMD)
	case FLAG_CPUID_3DNOW:		// 3DNow! (AMD)
		// extended CPUID(80000001) EDX (for Intel Standard)
		if( NULL == (pcpus = searchCPUIDdata(pci->pcpus, cntcpu, EXTCPUID+1)) ){ return FALSE; }
		bRet = pcpus->cpu.edx&(1<<(feature&FLAG_CPUID_BITSHIFT)) ? TRUE : FALSE;
		break;
	case FLAG_CPUID_LAHF4L:		// LAHF/SAHF for Long mode
	case FLAG_CPUID_CMPLegacy:  // CMPLegacy
	case FLAG_CPUID_LAHF:		// LAHF/SAHF

		// extended CPUID(80000001) ECX (for Intel Standard)
		if( NULL == (pcpus = searchCPUIDdata(pci->pcpus, cntcpu, EXTCPUID+1)) ){ return FALSE; }
		bRet = pcpus->cpu.ecx&(1<<(feature&FLAG_CPUID_BITSHIFT)) ? TRUE : FALSE;
		break;
	default:
		bRet = FALSE;
		break;
	}

	return bRet;
}



DWORD
GetProcessorNum(
	PCPUID_INFOR pci
	)
{
	if( NULL == pci ){ return 0; }
	return pci->dwProcessor;
}


// CPUVENDOR_INTEL, CPUVENDOR_AMD, and CPUVENDOR_UNKNOWN
DWORD
GetCPUVendor(
	PCPUID_INFOR pci
	)
{
	if( NULL == pci ){ return CPUVENDOR_UNKNOWN; }
	return pci->dwVendor;
}

PBYTE
GetVendorID(
	PCPUID_INFOR pci
	)
{
	if( NULL == pci ){ return NULL; }
	return pci->szVendorID;
}

PBYTE
GetVendorString(
	PCPUID_INFOR pci
	)
{
	if( NULL == pci ){ return NULL; }
	return pci->szBrand ;
}



BOOL
GetProcessorData(
	IN  PCPUID_INFOR pci,
	IN  DWORD cntcpu,
	IN  DWORD ind,
	OUT PDWORD pdw
	)
{
	PCPUIDS_DATA pcpus = NULL;

	if( NULL == pci || NULL == pdw ){ return FALSE; }

	switch(ind){
/* in:EAX=1, out=EBX */
	case CPUID_LOCAL_APIC:
		// in  : EAX = 1
		// out : EBX bit 31-24 : Processor local APIC physical ID (Valid for Pentium 4 and subsequent processors)
		if( NULL == (pcpus = searchCPUIDdata(pci->pcpus, cntcpu, 1)) ){ return FALSE; }
		*pdw = (pcpus->cpu.ebx&0xFF000000)>>24;
		break;
	case CPUID_LOGICAL_PROCESSOR_NUM:
		// in  : EAX = 1
		// out : EBX bit 23-16 : Count of logical processors (Valid only Hyper-Threading Technology flag is set)
		if( NULL == (pcpus = searchCPUIDdata(pci->pcpus, cntcpu, 1)) ){ return FALSE; }
		*pdw = (pcpus->cpu.ebx&0x00FF0000)>>16;
		break;
	case CPUID_BRAND_INDEX:
		// in  : EAX = 1
		// out : EBX bit 7-0 : Brand Index - Not supported if the value = 00h
		if( NULL == (pcpus = searchCPUIDdata(pci->pcpus, cntcpu, 1)) ){ return FALSE; }
		*pdw = (pcpus->cpu.ebx&0xFF);
		break;
/* in:EAX=1, out=EAX */
	case CPUID_FAMILY_CODE:
		// in  : EAX = 1
		// out : EAX bit 11-8 : Family Code
		if( NULL == (pcpus = searchCPUIDdata(pci->pcpus, cntcpu, 1)) ){ return FALSE; }
		*pdw = (pcpus->cpu.eax&0x0F00)>>8;
		break;
	case CPUID_MODEL_NUMBER:
		// in  : EAX = 1
		// out : EAX bit 7-4 : Model Number
		if( NULL == (pcpus = searchCPUIDdata(pci->pcpus, cntcpu, 1)) ){ return FALSE; }
		*pdw = (pcpus->cpu.eax&0xF0)>>4;
		break;
	case CPUID_STEPPING_ID:
		// in  : EAX = 1
		// out : EAX bit 3-0 : Stepping ID
		if( NULL == (pcpus = searchCPUIDdata(pci->pcpus, cntcpu, 1)) ){ return FALSE; }
		*pdw = (pcpus->cpu.eax&0xF);
		break;
	case CPUID_PROCESSOR_TYPE:
		// in  : EAX = 1
		// out : EAX bit 13-12 : Processor Type (00:Original OEM Processor, 01:OverDrive Processor, 10:Dual Processor, 11:Reserved.)
		if( NULL == (pcpus = searchCPUIDdata(pci->pcpus, cntcpu, 1)) ){ return FALSE; }
		*pdw = (pcpus->cpu.eax&0x3000)>>12;
		break;
	case CPUID_EXTENDED_MODEL:
		// in  : EAX = 1
		// out : EAX bit 19-16 : Extended Model
		if( NULL == (pcpus = searchCPUIDdata(pci->pcpus, cntcpu, 1)) ){ return FALSE; }
		*pdw = (pcpus->cpu.eax&0xF0000)>>16;
		break;
	case CPUID_EXTENDED_FAMILY:
		// in  : EAX = 1
		// out : EAX bit 27-20 : Extended Family
		if( NULL == (pcpus = searchCPUIDdata(pci->pcpus, cntcpu, 1)) ){ return FALSE; }
		*pdw = (pcpus->cpu.eax&0x0FF00000)>>20;
		break;
	default:
		return FALSE;
	}

	return TRUE;
}

DWORD
GetMultiCoreNum(
	IN  PCPUID_INFOR pci,
	IN  DWORD cntcpu
	)
{
	DWORD num = 0;
	PCPUIDS_DATA pcpus = NULL;

	if( NULL == pci ){ return 0; }

	switch(GetCPUVendor(pci)){
	case CPUVENDOR_INTEL:
		// in  : EAX = 4, EDX =0
		// out : EAX bit 31-26 : Number of processor cores on this die (Multicore)
		if( NULL == (pcpus = searchCPUIDdata(pci->pcpus, cntcpu, 4)) ){
			num = 0;
		}else{
			num = (pcpus->cpu.eax&0xFC000000)>>26;
			if( num != 0 ){
				num++;
			}
		}
		break;
	case CPUVENDOR_AMD:
		// in  : EAX = EXTCPUID+8)
		// out : ECX bit 7-0 : NC: number of logical CPU cores -1.
		if( NULL == (pcpus = searchCPUIDdata(pci->pcpus, cntcpu, EXTCPUID+8)) ){
			num = 0;
		}else{
			num = (pcpus->cpu.ecx&0xFF);
			if( num != 0 ){
				num++;
			}
		}
		break;
	default:
		num = 0;
		break;
	}
	return num;
}

DWORD
GetHyperThreadingNum(
	IN  PCPUID_INFOR pci,
	IN  DWORD cntcpu
	)
{
	DWORD num = 0;
	DWORD llp = 0;
	DWORD core = 0;
	PCPUIDS_DATA pcpus = NULL;

	if( NULL == pci ){ return 0; }

	switch(GetCPUVendor(pci)){
	case CPUVENDOR_INTEL:
		num = 0;
		// Count Logical Processor number (HTT+core)
		if( !GetProcessorData(pci,cntcpu,CPUID_LOGICAL_PROCESSOR_NUM, &llp) ){
			break;
		}
		// HTT enable? (if Logical Processor is 1, then HTT is NOT working.)
		if( !IsFeature(pci,cntcpu,FLAG_CPUID_HTT) || llp<2 ){
			break;
		}
		// Get Multicore number
		core = GetMultiCoreNum(pci,cntcpu);
		if( 0 != core ){
			if( (llp/core) < 2 ){
				break;
			}
		}
		num = llp;
		break;
	default:
		num = 0;
		break;
	}

	return num;
}


BOOL
TotalProcessorNum(
	IN  PCPUID_INFOR	pci,
	OUT PDWORD			pdwProcessor,		// total processor number
	OUT	PDWORD			pdwPhysical,		// Physical processor number (socket)
	OUT PDWORD			pdwCore,			// Multicore number
	OUT PDWORD			pdwHT				// Hyper-Threading number
	)
{
	DWORD core = 0, htt = 0, num = 0, llp = 0;

	if( NULL == pci ){ return FALSE; }	

	num  = GetProcessorNum(pci);
	htt  = GetHyperThreadingNum(pci,0);
	core = GetMultiCoreNum(pci,0);
	// Count Logical Processor number (HTT+core)
	if( !GetProcessorData(pci,0,CPUID_LOGICAL_PROCESSOR_NUM, &llp) ){
		return FALSE;
	}

	if( NULL != pdwProcessor ){
		*pdwProcessor = num;
	}
	if( NULL != pdwCore ){
		*pdwCore = core;
	}
	if( NULL != pdwHT ){
		*pdwHT = htt;
	}
	if( NULL != pdwPhysical ){
		if( 0 == llp ){ llp = 1; }
		*pdwPhysical = num / llp;
	}
	return TRUE;
}

PCPUID_DATA
FindFirstCPUID(
	PCPUID_INFOR pci
	)
{
	if( NULL == pci || NULL == pci->pcpus ){ return NULL; }	
	pci->pnext = pci->pcpus->pnext;
	return &(pci->pcpus->cpu);
}

PCPUID_DATA
FindNextCPUID(
	PCPUID_INFOR pci
	)
{
	PCPUIDS_DATA p = NULL;
	if( NULL == pci ){ return NULL; }
	if( NULL == pci->pnext ){ return NULL; }
	p = pci->pnext;
	pci->pnext = p->pnext;
	return &(p->cpu);
}


#ifdef _SAMPLE_BUILD
int
main(void)
{
	PCPUID_INFOR pci = NULL;
	DWORD dwProc, dwPhy, dwCore, dwHTT;
	PBYTE pBrand = NULL;
#ifdef __CHECK_MEMORY_LEAK_
	_CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
#endif 
	pci = OpenCPUIDinfo();
	if( NULL==pci ){ return -1; }
	
	pBrand = GetVendorString(pci);
	if( NULL != pBrand ){
		printf("Brand Srting = [%s]\n", pBrand);
	}
	if( TotalProcessorNum(pci,&dwProc,&dwPhy,&dwCore,&dwHTT) ){
		printf("Total processor            = [%d]\n", dwProc );
		printf("Physical  Processor number = [%d]\n", dwPhy  );
		printf("Multicore Processor number = [%d]\n", dwCore );
		printf("HTT       Processor number = [%d]\n", dwHTT  );
	}

	// close
	CloseCPUIDinfo(pci);

#ifdef __CHECK_MEMORY_LEAK_
	_CrtDumpMemoryLeaks();
#endif

	return 0;
}
#endif _SAMPLE_BUILD


